// Copyright (c) 2014 The ShadowCoin developers
// Copyright (c) 2017-2020 The Particl Core developers
// Distributed under the MIT/X11 software license, see the accompanying
// file license.txt or http://www.opensource.org/licenses/mit-license.php.

#include <key/stealth.h>

#include <key_io.h>
#include <key/keyutil.h>
#include <pubkey.h>
#include <random.h>
#include <script/script.h>
#include <smsg/crypter.h>
#include <serialize.h>

#include <support/allocators/secure.h>

#include <cmath>
#include <secp256k1.h>
#include <secp256k1_ecdh.h>

secp256k1_context *secp256k1_ctx_stealth = nullptr;

bool CStealthAddress::SetEncoded(const std::string &encodedAddress)
{
    std::vector<uint8_t> raw;

    if (!DecodeBase58(encodedAddress, raw, MAX_STEALTH_RAW_SIZE)) {
        LogPrint(BCLog::HDWALLET, "%s: DecodeBase58 failed.\n", __func__);
        return false;
    }

    if (!VerifyChecksum(raw)) {
        LogPrint(BCLog::HDWALLET, "%s: verify_checksum failed.\n", __func__);
        return false;
    }

    if (raw.size() < MIN_STEALTH_RAW_SIZE + 5) {
        LogPrint(BCLog::HDWALLET, "%s: too few bytes provided.\n", __func__);
        return false;
    }

    uint8_t *p = &raw[0];
    uint8_t version = *p++;

    if (version != Params().Base58Prefix(CChainParams::STEALTH_ADDRESS)[0]) {
        LogPrintf("%s: version mismatch 0x%x != 0x%x.\n", __func__, version, Params().Base58Prefix(CChainParams::STEALTH_ADDRESS)[0]);
        return false;
    }

    return 0 == FromRaw(p, raw.size()-1);
};

int CStealthAddress::FromRaw(const uint8_t *p, size_t nSize)
{
    if (nSize < MIN_STEALTH_RAW_SIZE) {
        return 1;
    }
    options = *p++;

    scan_pubkey.resize(EC_COMPRESSED_SIZE);
    memcpy(&scan_pubkey[0], p, EC_COMPRESSED_SIZE);
    p += EC_COMPRESSED_SIZE;
    uint8_t spend_pubkeys = *p++;
    if (spend_pubkeys < 1) {
        return 1;
    }

    if (nSize < MIN_STEALTH_RAW_SIZE + EC_COMPRESSED_SIZE * (spend_pubkeys-1)) {
        return 1;
    }

    spend_pubkey.resize(EC_COMPRESSED_SIZE * spend_pubkeys);
    memcpy(&spend_pubkey[0], p, EC_COMPRESSED_SIZE * spend_pubkeys);
    p += EC_COMPRESSED_SIZE * spend_pubkeys;
    number_signatures = *p++;
    prefix.number_bits = *p++;
    prefix.bitfield = 0;
    size_t nPrefixBytes = std::ceil((float)prefix.number_bits / 8.0);

    if (nPrefixBytes > MAX_STEALTH_PREFIX_BYTES ||
        nSize < MIN_STEALTH_RAW_SIZE + EC_COMPRESSED_SIZE * (spend_pubkeys-1) + nPrefixBytes) {
        return 1;
    }

    if (nPrefixBytes) {
        memcpy(&prefix.bitfield, p, nPrefixBytes);
        prefix.bitfield = le32toh(prefix.bitfield);
    }

    return 0;
}

int CStealthAddress::ToRaw(std::vector<uint8_t> &raw) const
{
    // https://wiki.unsystem.net/index.php/DarkWallet/Stealth#Address_format
    // [version] [options] [scan_key] [N] ... [Nsigs] [prefix_length] ...

    size_t nPrefixBytes = std::ceil((float)prefix.number_bits / 8.0);
    size_t nPkSpend = spend_pubkey.size() / EC_COMPRESSED_SIZE;
    if (scan_pubkey.size() != EC_COMPRESSED_SIZE
        || spend_pubkey.size() < EC_COMPRESSED_SIZE
        || spend_pubkey.size() % EC_COMPRESSED_SIZE != 0
        || nPkSpend > MAX_STEALTH_SPEND_KEYS
        || nPrefixBytes > MAX_STEALTH_PREFIX_BYTES) {
        LogPrintf("%s: sanity checks failed.\n", __func__);
        return 1;
    }

    raw.resize(MIN_STEALTH_RAW_SIZE + EC_COMPRESSED_SIZE * (nPkSpend-1) + nPrefixBytes);

    int o = 0;
    raw[o] = options; o++;
    memcpy(&raw[o], &scan_pubkey[0], EC_COMPRESSED_SIZE); o += EC_COMPRESSED_SIZE;
    raw[o] = nPkSpend; o++;
    memcpy(&raw[o], &spend_pubkey[0], EC_COMPRESSED_SIZE * nPkSpend); o += EC_COMPRESSED_SIZE * nPkSpend;
    raw[o] = number_signatures; o++;
    raw[o] = prefix.number_bits; o++;
    if (nPrefixBytes) {
        uint32_t tmp = htole32(prefix.bitfield);
        memcpy(&raw[o], &tmp, nPrefixBytes); o += nPrefixBytes;
    }

    return 0;
};

std::string CStealthAddress::Encoded(bool fBech32) const
{
    return CBitcoinAddress(*this, fBech32).ToString();
};

int CStealthAddress::SetScanPubKey(CPubKey pk)
{
    scan_pubkey.resize(pk.size());
    memcpy(&scan_pubkey[0], pk.begin(), pk.size());
    return 0;
};

CKeyID CStealthAddress::GetSpendKeyID() const
{
    return CKeyID(Hash160(spend_pubkey));
};

CKeyID CStealthAddress::GetScanKeyID() const
{
    return CKeyID(Hash160(scan_pubkey));
};

int SecretToPublicKey(const CKey &secret, ec_point &out)
{
    // Public key = private * G

    CPubKey pkTemp = secret.GetPubKey();
    out.resize(EC_COMPRESSED_SIZE);
    memcpy(&out[0], pkTemp.begin(), EC_COMPRESSED_SIZE);
    return 0;
};

int SetPublicKey(const CPubKey &pk, ec_point &out)
{
    out.resize(EC_COMPRESSED_SIZE);
    memcpy(&out[0], pk.begin(), EC_COMPRESSED_SIZE);
    return 0;
};

int StealthShared(const CKey &secret, const ec_point &pubkey, CKey &sharedSOut)
{
    if (pubkey.size() != EC_COMPRESSED_SIZE) {
        return errorN(1, "%s: sanity checks failed.", __func__);
    }

    secp256k1_pubkey Q;
    if (!secp256k1_ec_pubkey_parse(secp256k1_ctx_stealth, &Q, &pubkey[0], EC_COMPRESSED_SIZE)) {
        return errorN(1, "%s: secp256k1_ec_pubkey_parse Q failed.", __func__);
    }

    // H(eQ)
    if (!secp256k1_ecdh(secp256k1_ctx_stealth, sharedSOut.begin_nc(), &Q, secret.begin(), nullptr, nullptr)) {
        return errorN(1, "%s: secp256k1_ctx_stealth failed.", __func__);
    }

    return 0;
};

int StealthSecret(const CKey &secret, const ec_point &pubkey, const ec_point &pkSpend, CKey &sharedSOut, ec_point &pkOut)
{
    /*
    send:
        secret = ephem_secret, pubkey = scan_pubkey

    receive:
        secret = scan_secret, pubkey = ephem_pubkey
        c = H(dP)

    Q = public scan key (EC point, 33 bytes)
    d = private scan key (integer, 32 bytes)
    R = public spend key
    f = private spend key

    Q = dG
    R = fG

    Sender (has Q and R, not d or f):

    P = eG

    c = H(eQ) = H(dP)
    R' = R + cG


    Recipient gets R' and P

    test 0 and infinity?
    */

    if (pubkey.size() != EC_COMPRESSED_SIZE
        || pkSpend.size() != EC_COMPRESSED_SIZE) {
        return errorN(1, "%s: sanity checks failed.", __func__);
    }

    secp256k1_pubkey Q, R;
    if (!secp256k1_ec_pubkey_parse(secp256k1_ctx_stealth, &Q, &pubkey[0], EC_COMPRESSED_SIZE)) {
        return errorN(1, "%s: secp256k1_ec_pubkey_parse Q failed.", __func__);
    }
    if (!secp256k1_ec_pubkey_parse(secp256k1_ctx_stealth, &R, &pkSpend[0], EC_COMPRESSED_SIZE)) {
        return errorN(1, "%s: secp256k1_ec_pubkey_parse R failed.", __func__);
    }

    // H(eQ)
    if (!secp256k1_ecdh(secp256k1_ctx_stealth, sharedSOut.begin_nc(), &Q, secret.begin(), nullptr, nullptr)) {
        return errorN(1, "%s: secp256k1_ctx_stealth failed.", __func__);
    }

    // C = sharedSOut * G
    // R' = R + C
    if (!secp256k1_ec_pubkey_tweak_add(secp256k1_ctx_stealth, &R, sharedSOut.begin())) {
        return errorN(1, "%s: secp256k1_ec_pubkey_tweak_add failed.", __func__); // Start again with a new ephemeral key
    }

    try {
        pkOut.resize(EC_COMPRESSED_SIZE);
    } catch (std::exception &e) {
        return errorN(8, "%s: pkOut.resize %u threw: %s.", __func__, EC_COMPRESSED_SIZE);
    };

    size_t len = 33;
    secp256k1_ec_pubkey_serialize(secp256k1_ctx_stealth, &pkOut[0], &len, &R, SECP256K1_EC_COMPRESSED); // Returns: 1 always.

    return 0;
};


int StealthSecretSpend(const CKey &scanSecret, const ec_point &ephemPubkey, const CKey &spendSecret, CKey &secretOut)
{
    /*
    c  = H(dP)
    R' = R + cG     [without decrypting wallet]
       = (f + c)G   [after decryption of wallet]
    */

    if (ephemPubkey.size() != EC_COMPRESSED_SIZE) {
        return errorN(1, "%s: sanity checks failed.", __func__);
    }

    secp256k1_pubkey P;
    if (!secp256k1_ec_pubkey_parse(secp256k1_ctx_stealth, &P, &ephemPubkey[0], EC_COMPRESSED_SIZE)) {
        return errorN(1, "%s: secp256k1_ec_pubkey_parse P failed.", __func__);
    }

    uint8_t tmp32[32];
    // H(dP)
    if (!secp256k1_ecdh(secp256k1_ctx_stealth, tmp32, &P, scanSecret.begin(), nullptr, nullptr)) {
        return errorN(1, "%s: secp256k1_ctx_stealth failed.", __func__);
    }

    secretOut = spendSecret;
    if (!secp256k1_ec_privkey_tweak_add(secp256k1_ctx_stealth, secretOut.begin_nc(), tmp32)) {
        return errorN(1, "%s: secp256k1_ec_privkey_tweak_add failed.", __func__);
    }

    return 0;
};


int StealthSharedToSecretSpend(const CKey &sharedS, const CKey &spendSecret, CKey &secretOut)
{
    secretOut = spendSecret;
    if (!secp256k1_ec_privkey_tweak_add(secp256k1_ctx_stealth, secretOut.begin_nc(), sharedS.begin())) {
        return errorN(1, "%s: secp256k1_ec_privkey_tweak_add failed.", __func__);
    }

    if (!secp256k1_ec_seckey_verify(secp256k1_ctx_stealth, secretOut.begin())) { // necessary?
        return errorN(1, "%s: secp256k1_ec_seckey_verify failed.", __func__);
    }

    return 0;
};

int StealthSharedToPublicKey(const ec_point &pkSpend, const CKey &sharedS, ec_point &pkOut)
{
    if (pkSpend.size() != EC_COMPRESSED_SIZE) {
        return errorN(1, "%s: sanity checks failed.", __func__);
    }

    secp256k1_pubkey R;
    if (!secp256k1_ec_pubkey_parse(secp256k1_ctx_stealth, &R, &pkSpend[0], EC_COMPRESSED_SIZE)) {
        return errorN(1, "%s: secp256k1_ec_pubkey_parse R failed.", __func__);
    }

    if (!secp256k1_ec_pubkey_tweak_add(secp256k1_ctx_stealth, &R, sharedS.begin())) {
        return errorN(1, "%s: secp256k1_ec_pubkey_tweak_add failed.", __func__);
    }

    try {
        pkOut.resize(EC_COMPRESSED_SIZE);
    } catch (std::exception &e) {
        return errorN(8, "%s: pkOut.resize %u threw: %s.", __func__, EC_COMPRESSED_SIZE);
    };

    size_t len = 33;
    secp256k1_ec_pubkey_serialize(secp256k1_ctx_stealth, &pkOut[0], &len, &R, SECP256K1_EC_COMPRESSED); // Returns: 1 always.

    return 0;
};

bool IsStealthAddress(const std::string &encodedAddress)
{
    std::vector<uint8_t> raw;

    if (!DecodeBase58(encodedAddress, raw, MAX_STEALTH_RAW_SIZE)) {
        //LogPrintf("IsStealthAddress DecodeBase58 failed.\n");
        return false;
    }

    if (!VerifyChecksum(raw)) {
        //LogPrintf("IsStealthAddress verify_checksum failed.\n");
        return false;
    }

    if (raw.size() < MIN_STEALTH_RAW_SIZE + 5) {
        //LogPrintf("IsStealthAddress too few bytes provided.\n");
        return false;
    }

    uint8_t *p = &raw[0];
    uint8_t version = *p++;

    if (version != Params().Base58Prefix(CChainParams::STEALTH_ADDRESS)[0]) {
        //LogPrintf("IsStealthAddress version mismatch 0x%x != 0x%x.\n", version, stealth_version_byte);
        return false;
    }

    return true;
};

uint32_t FillStealthPrefix(uint8_t nBits, uint32_t nBitfield)
{
    uint32_t prefix, mask = SetStealthMask(nBits);
    GetStrongRandBytes((uint8_t*) &prefix, 4);

    prefix &= (~mask);
    prefix |= nBitfield & mask;
    return prefix;
};

bool ExtractStealthPrefix(const char *pPrefix, uint32_t &nPrefix)
{
    int base = 10;
    size_t len = strlen(pPrefix);
    if (len > 2
        && pPrefix[0] == '0') {
        if (pPrefix[1] == 'b') {
            pPrefix += 2;
            base = 2;
        } else
        if (pPrefix[1] == 'x' || pPrefix[1] == 'X') {
            pPrefix += 2;
            base = 16;
        }
    }

    char *pend;
    errno = 0;
    nPrefix = strtol(pPrefix, &pend, base);

    if (errno != 0 || !pend || *pend != '\0') {
        return error("%s strtol failed.", __func__);
    }
    return true;
};

int MakeStealthData(const std::string &sNarration, stealth_prefix prefix, const CKey &sShared, const CPubKey &pkEphem,
    std::vector<uint8_t> &vData, uint32_t &nStealthPrefix, std::string &sError)
{
    std::vector<uint8_t> vchNarr;
    if (sNarration.length() > 0) {
        SecMsgCrypter crypter;
        crypter.SetKey(sShared.begin(), pkEphem.begin());

        if (!crypter.Encrypt((uint8_t*)sNarration.data(), sNarration.length(), vchNarr)) {
            return errorN(1, sError, __func__, "Narration encryption failed.");
        }

        if (vchNarr.size() > MAX_STEALTH_NARRATION_SIZE) {
            return errorN(1, sError, __func__, "Encrypted narration is too long.");
        }
    }

    vData.resize(34
        + (prefix.number_bits > 0 ? 5 : 0)
        + (vchNarr.size() + (vchNarr.size() > 0 ? 1 : 0)));

    size_t o = 0;
    vData[o++] = DO_STEALTH;
    memcpy(&vData[o], pkEphem.begin(), 33);
    o += 33;

    if (prefix.number_bits > 0) {
        vData[o++] = DO_STEALTH_PREFIX;
        nStealthPrefix = FillStealthPrefix(prefix.number_bits, prefix.bitfield);
        uint32_t tmp = htole32(nStealthPrefix);
        memcpy(&vData[o], &tmp, 4);
        o+=4;
    }

    if (vchNarr.size() > 0) {
        vData[o++] = DO_NARR_CRYPT;
        memcpy(&vData[o], vchNarr.data(), vchNarr.size());
        o += vchNarr.size();
    }

    return 0;
};

int PrepareStealthOutput(const CStealthAddress &sx, const std::string &sNarration,
    CScript &scriptPubKey, std::vector<uint8_t> &vData, std::string &sError)
{
    CKey sShared, sEphem;
    ec_point pkSendTo;
    int k, nTries = 24;
    for (k = 0; k < nTries; ++k) { // if StealthSecret fails try again with new ephem key
        sEphem.MakeNewKey(true);
        if (StealthSecret(sEphem, sx.scan_pubkey, sx.spend_pubkey, sShared, pkSendTo) == 0)
            break;
    }
    if (k >= nTries) {
        return errorN(1, sError, __func__, "Could not generate receiving public key.");
    }
    CPubKey pkEphem = sEphem.GetPubKey();
    scriptPubKey = GetScriptForDestination(PKHash(CPubKey(pkSendTo)));

    uint32_t nStealthPrefix;
    if (0 != MakeStealthData(sNarration, sx.prefix, sShared, pkEphem, vData, nStealthPrefix, sError)) {
        return 1;
    }
    return 0;
};

void ECC_Start_Stealth()
{
    assert(secp256k1_ctx_stealth == nullptr);

    secp256k1_context *ctx = secp256k1_context_create(SECP256K1_CONTEXT_SIGN | SECP256K1_CONTEXT_VERIFY);
    assert(ctx != nullptr);

    {
        // Pass in a random blinding seed to the secp256k1 context.
        std::vector<unsigned char, secure_allocator<unsigned char>> vseed(32);
        GetRandBytes(vseed.data(), 32);
        bool ret = secp256k1_context_randomize(ctx, vseed.data());
        assert(ret);
    }

    secp256k1_ctx_stealth = ctx;
};

void ECC_Stop_Stealth()
{
    secp256k1_context *ctx = secp256k1_ctx_stealth;
    secp256k1_ctx_stealth = nullptr;

    if (ctx) {
        secp256k1_context_destroy(ctx);
    }
};
