// Copyright (c) 2009-2010 Satoshi Nakamoto
// Copyright (c) 2009-2020 The Bitcoin Core developers
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

#include <txdb.h>

#include <node/ui_interface.h>
#include <pow.h>
#include <random.h>
#include <shutdown.h>
#include <uint256.h>
#include <util/memory.h>
#include <util/system.h>
#include <util/translation.h>
#include <util/vector.h>

#include <validation.h>
#include <insight/insight.h>
#include <chainparams.h>

#include <stdint.h>

static const char DB_COIN = 'C';
//static const char DB_COINS = 'c';
static const char DB_BLOCK_FILES = 'f';
//static const char DB_TXINDEX = 't';
static const char DB_ADDRESSINDEX = 'a';
static const char DB_ADDRESSUNSPENTINDEX = 'u';
static const char DB_TIMESTAMPINDEX = 's';
static const char DB_BLOCKHASHINDEX = 'z';
static const char DB_SPENTINDEX = 'p';
static const char DB_BALANCESINDEX = 'i';
//static const char DB_TXINDEX_BLOCK = 'T';
static const char DB_BLOCK_INDEX = 'b';

static const char DB_BEST_BLOCK = 'B';
static const char DB_HEAD_BLOCKS = 'H';
static const char DB_FLAG = 'F';
static const char DB_REINDEX_FLAG = 'R';
static const char DB_LAST_BLOCK = 'l';

/*
static const char DB_RCTOUTPUT = 'A';
static const char DB_RCTOUTPUT_LINK = 'L';
static const char DB_RCTKEYIMAGE = 'K';
static const char DB_SPENTCACHE = 'S';
*/

namespace {

struct CoinEntry {
    COutPoint* outpoint;
    char key;
    explicit CoinEntry(const COutPoint* ptr) : outpoint(const_cast<COutPoint*>(ptr)), key(DB_COIN)  {}

    SERIALIZE_METHODS(CoinEntry, obj) { READWRITE(obj.key, obj.outpoint->hash, VARINT(obj.outpoint->n)); }
};

}

CCoinsViewDB::CCoinsViewDB(fs::path ldb_path, size_t nCacheSize, bool fMemory, bool fWipe) :
    m_db(MakeUnique<CDBWrapper>(ldb_path, nCacheSize, fMemory, fWipe, true, false, 64)),
    m_ldb_path(ldb_path),
    m_is_memory(fMemory) { }

void CCoinsViewDB::ResizeCache(size_t new_cache_size)
{
    // Have to do a reset first to get the original `m_db` state to release its
    // filesystem lock.
    m_db.reset();
    m_db = MakeUnique<CDBWrapper>(
        m_ldb_path, new_cache_size, m_is_memory, /*fWipe*/ false, /*obfuscate*/ true);
}

bool CCoinsViewDB::GetCoin(const COutPoint &outpoint, Coin &coin) const {
    return m_db->Read(CoinEntry(&outpoint), coin);
}

bool CCoinsViewDB::HaveCoin(const COutPoint &outpoint) const {
    return m_db->Exists(CoinEntry(&outpoint));
}

uint256 CCoinsViewDB::GetBestBlock() const {
    uint256 hashBestChain;
    if (!m_db->Read(DB_BEST_BLOCK, hashBestChain))
        return uint256();
    return hashBestChain;
}

std::vector<uint256> CCoinsViewDB::GetHeadBlocks() const {
    std::vector<uint256> vhashHeadBlocks;
    if (!m_db->Read(DB_HEAD_BLOCKS, vhashHeadBlocks)) {
        return std::vector<uint256>();
    }
    return vhashHeadBlocks;
}

bool CCoinsViewDB::BatchWrite(CCoinsMap &mapCoins, const uint256 &hashBlock) {
    CDBBatch batch(*m_db);
    size_t count = 0;
    size_t changed = 0;
    size_t batch_size = (size_t)gArgs.GetArg("-dbbatchsize", nDefaultDbBatchSize);
    int crash_simulate = gArgs.GetArg("-dbcrashratio", 0);
    assert(!hashBlock.IsNull());

    uint256 old_tip = GetBestBlock();
    if (old_tip.IsNull()) {
        // We may be in the middle of replaying.
        std::vector<uint256> old_heads = GetHeadBlocks();
        if (old_heads.size() == 2) {
            assert(old_heads[0] == hashBlock);
            old_tip = old_heads[1];
        }
    }

    // In the first batch, mark the database as being in the middle of a
    // transition from old_tip to hashBlock.
    // A vector is used for future extensibility, as we may want to support
    // interrupting after partial writes from multiple independent reorgs.
    batch.Erase(DB_BEST_BLOCK);
    batch.Write(DB_HEAD_BLOCKS, Vector(hashBlock, old_tip));

    for (CCoinsMap::iterator it = mapCoins.begin(); it != mapCoins.end();) {
        if (it->second.flags & CCoinsCacheEntry::DIRTY) {
            CoinEntry entry(&it->first);
            if (it->second.coin.IsSpent())
                batch.Erase(entry);
            else
                batch.Write(entry, it->second.coin);
            changed++;
        }
        count++;
        CCoinsMap::iterator itOld = it++;
        mapCoins.erase(itOld);
        if (batch.SizeEstimate() > batch_size) {
            LogPrint(BCLog::COINDB, "Writing partial batch of %.2f MiB\n", batch.SizeEstimate() * (1.0 / 1048576.0));
            m_db->WriteBatch(batch);
            batch.Clear();
            if (crash_simulate) {
                static FastRandomContext rng;
                if (rng.randrange(crash_simulate) == 0) {
                    LogPrintf("Simulating a crash. Goodbye.\n");
                    _Exit(0);
                }
            }
        }
    }

    // In the last batch, mark the database as consistent with hashBlock again.
    batch.Erase(DB_HEAD_BLOCKS);
    batch.Write(DB_BEST_BLOCK, hashBlock);

    LogPrint(BCLog::COINDB, "Writing final batch of %.2f MiB\n", batch.SizeEstimate() * (1.0 / 1048576.0));
    bool ret = m_db->WriteBatch(batch);
    LogPrint(BCLog::COINDB, "Committed %u changed transaction outputs (out of %u) to coin database...\n", (unsigned int)changed, (unsigned int)count);
    return ret;
}

size_t CCoinsViewDB::EstimateSize() const
{
    return m_db->EstimateSize(DB_COIN, (char)(DB_COIN+1));
}

CBlockTreeDB::CBlockTreeDB(size_t nCacheSize, bool fMemory, bool fWipe, bool compression, int maxOpenFiles) : CDBWrapper(GetDataDir() / "blocks" / "index", nCacheSize, fMemory, fWipe, false, compression, maxOpenFiles) {
}

bool CBlockTreeDB::ReadBlockFileInfo(int nFile, CBlockFileInfo &info) {
    return Read(std::make_pair(DB_BLOCK_FILES, nFile), info);
}

bool CBlockTreeDB::WriteReindexing(bool fReindexing) {
    if (fReindexing)
        return Write(DB_REINDEX_FLAG, '1');
    else
        return Erase(DB_REINDEX_FLAG);
}

void CBlockTreeDB::ReadReindexing(bool &fReindexing) {
    fReindexing = Exists(DB_REINDEX_FLAG);
}

bool CBlockTreeDB::ReadLastBlockFile(int &nFile) {
    return Read(DB_LAST_BLOCK, nFile);
}

CCoinsViewCursor *CCoinsViewDB::Cursor() const
{
    CCoinsViewDBCursor *i = new CCoinsViewDBCursor(const_cast<CDBWrapper&>(*m_db).NewIterator(), GetBestBlock());
    /* It seems that there are no "const iterators" for LevelDB.  Since we
       only need read operations on it, use a const-cast to get around
       that restriction.  */
    i->pcursor->Seek(DB_COIN);
    // Cache key of first record
    if (i->pcursor->Valid()) {
        CoinEntry entry(&i->keyTmp.second);
        i->pcursor->GetKey(entry);
        i->keyTmp.first = entry.key;
    } else {
        i->keyTmp.first = 0; // Make sure Valid() and GetKey() return false
    }
    return i;
}

bool CCoinsViewDBCursor::GetKey(COutPoint &key) const
{
    // Return cached key
    if (keyTmp.first == DB_COIN) {
        key = keyTmp.second;
        return true;
    }
    return false;
}

bool CCoinsViewDBCursor::GetValue(Coin &coin) const
{
    return pcursor->GetValue(coin);
}

unsigned int CCoinsViewDBCursor::GetValueSize() const
{
    return pcursor->GetValueSize();
}

bool CCoinsViewDBCursor::Valid() const
{
    return keyTmp.first == DB_COIN;
}

void CCoinsViewDBCursor::Next()
{
    pcursor->Next();
    CoinEntry entry(&keyTmp.second);
    if (!pcursor->Valid() || !pcursor->GetKey(entry)) {
        keyTmp.first = 0; // Invalidate cached key after last record so that Valid() and GetKey() return false
    } else {
        keyTmp.first = entry.key;
    }
}

bool CBlockTreeDB::WriteBatchSync(const std::vector<std::pair<int, const CBlockFileInfo*> >& fileInfo, int nLastFile, const std::vector<const CBlockIndex*>& blockinfo) {
    CDBBatch batch(*this);
    for (std::vector<std::pair<int, const CBlockFileInfo*> >::const_iterator it=fileInfo.begin(); it != fileInfo.end(); it++) {
        batch.Write(std::make_pair(DB_BLOCK_FILES, it->first), *it->second);
    }
    batch.Write(DB_LAST_BLOCK, nLastFile);
    for (std::vector<const CBlockIndex*>::const_iterator it=blockinfo.begin(); it != blockinfo.end(); it++) {
        batch.Write(std::make_pair(DB_BLOCK_INDEX, (*it)->GetBlockHash()), CDiskBlockIndex(*it));
    }
    return WriteBatch(batch, true);
}

bool CBlockTreeDB::ReadSpentIndex(const CSpentIndexKey &key, CSpentIndexValue &value) {
    return Read(std::make_pair(DB_SPENTINDEX, key), value);
}

bool CBlockTreeDB::UpdateSpentIndex(const std::vector<std::pair<CSpentIndexKey, CSpentIndexValue> >&vect) {
    CDBBatch batch(*this);
    for (std::vector<std::pair<CSpentIndexKey,CSpentIndexValue> >::const_iterator it=vect.begin(); it!=vect.end(); it++) {
        if (it->second.IsNull()) {
            batch.Erase(std::make_pair(DB_SPENTINDEX, it->first));
        } else {
            batch.Write(std::make_pair(DB_SPENTINDEX, it->first), it->second);
        }
    }
    return WriteBatch(batch);
}

bool CBlockTreeDB::UpdateAddressUnspentIndex(const std::vector<std::pair<CAddressUnspentKey, CAddressUnspentValue > >&vect) {
    CDBBatch batch(*this);
    for (std::vector<std::pair<CAddressUnspentKey, CAddressUnspentValue> >::const_iterator it=vect.begin(); it!=vect.end(); it++) {
        if (it->second.IsNull()) {
            batch.Erase(std::make_pair(DB_ADDRESSUNSPENTINDEX, it->first));
        } else {
            batch.Write(std::make_pair(DB_ADDRESSUNSPENTINDEX, it->first), it->second);
        }
    }
    return WriteBatch(batch);
}

bool CBlockTreeDB::ReadAddressUnspentIndex(uint256 addressHash, int type,
                                           std::vector<std::pair<CAddressUnspentKey, CAddressUnspentValue> > &unspentOutputs) {
    const std::unique_ptr<CDBIterator> pcursor(NewIterator());

    pcursor->Seek(std::make_pair(DB_ADDRESSUNSPENTINDEX, CAddressIndexIteratorKey(type, addressHash)));

    while (pcursor->Valid()) {
        if (ShutdownRequested()) return false;
        std::pair<char, CAddressUnspentKey> key;
        if (pcursor->GetKey(key) && key.first == DB_ADDRESSUNSPENTINDEX && key.second.hashBytes == addressHash) {
            CAddressUnspentValue nValue;
            if (pcursor->GetValue(nValue)) {
                unspentOutputs.push_back(std::make_pair(key.second, nValue));
                pcursor->Next();
            } else {
                return error("failed to get address unspent value");
            }
        } else {
            break;
        }
    }

    return true;
}

bool CBlockTreeDB::WriteAddressIndex(const std::vector<std::pair<CAddressIndexKey, CAmount > >&vect) {
    CDBBatch batch(*this);
    for (std::vector<std::pair<CAddressIndexKey, CAmount> >::const_iterator it=vect.begin(); it!=vect.end(); it++)
        batch.Write(std::make_pair(DB_ADDRESSINDEX, it->first), it->second);
    return WriteBatch(batch);
}

bool CBlockTreeDB::EraseAddressIndex(const std::vector<std::pair<CAddressIndexKey, CAmount > >&vect) {
    CDBBatch batch(*this);
    for (std::vector<std::pair<CAddressIndexKey, CAmount> >::const_iterator it=vect.begin(); it!=vect.end(); it++)
        batch.Erase(std::make_pair(DB_ADDRESSINDEX, it->first));
    return WriteBatch(batch);
}

bool CBlockTreeDB::ReadAddressIndex(uint256 addressHash, int type,
                                    std::vector<std::pair<CAddressIndexKey, CAmount> > &addressIndex,
                                    int start, int end) {
    const std::unique_ptr<CDBIterator> pcursor(NewIterator());

    if (start > 0 && end > 0) {
        pcursor->Seek(std::make_pair(DB_ADDRESSINDEX, CAddressIndexIteratorHeightKey(type, addressHash, start)));
    } else {
        pcursor->Seek(std::make_pair(DB_ADDRESSINDEX, CAddressIndexIteratorKey(type, addressHash)));
    }

    while (pcursor->Valid()) {
        if (ShutdownRequested()) return false;
        std::pair<char, CAddressIndexKey> key;
        if (pcursor->GetKey(key) && key.first == DB_ADDRESSINDEX && key.second.hashBytes == addressHash) {
            if (end > 0 && key.second.blockHeight > end) {
                break;
            }
            CAmount nValue;
            if (pcursor->GetValue(nValue)) {
                addressIndex.push_back(std::make_pair(key.second, nValue));
                pcursor->Next();
            } else {
                return error("failed to get address index value");
            }
        } else {
            break;
        }
    }

    return true;
}

bool CBlockTreeDB::WriteTimestampIndex(const CTimestampIndexKey &timestampIndex)
{
    CDBBatch batch(*this);
    batch.Write(std::make_pair(DB_TIMESTAMPINDEX, timestampIndex), 0);
    return WriteBatch(batch);
}

bool CBlockTreeDB::ReadTimestampIndex(const unsigned int &high, const unsigned int &low, const bool fActiveOnly, std::vector<std::pair<uint256, unsigned int> > &hashes)
{
    const std::unique_ptr<CDBIterator> pcursor(NewIterator());

    pcursor->Seek(std::make_pair(DB_TIMESTAMPINDEX, CTimestampIndexIteratorKey(low)));

    if (fActiveOnly) {
        while (pcursor->Valid()) {
            if (ShutdownRequested()) return false;
            std::pair<char, CTimestampIndexKey> key;
            if (pcursor->GetKey(key) && key.first == DB_TIMESTAMPINDEX && key.second.timestamp < high) {
                if (HashOnchainActive(key.second.blockHash)) {
                    hashes.push_back(std::make_pair(key.second.blockHash, key.second.timestamp));
                }
                pcursor->Next();
            } else {
                break;
            }
        }
    } else {
        while (pcursor->Valid()) {
            if (ShutdownRequested()) return false;
            std::pair<char, CTimestampIndexKey> key;
            if (pcursor->GetKey(key) && key.first == DB_TIMESTAMPINDEX && key.second.timestamp < high) {
                hashes.push_back(std::make_pair(key.second.blockHash, key.second.timestamp));
                pcursor->Next();
            } else {
                break;
            }
        }
    }

    return true;
}

bool CBlockTreeDB::WriteTimestampBlockIndex(const CTimestampBlockIndexKey &blockhashIndex, const CTimestampBlockIndexValue &logicalts) {
    CDBBatch batch(*this);
    batch.Write(std::make_pair(DB_BLOCKHASHINDEX, blockhashIndex), logicalts);
    return WriteBatch(batch);
}

bool CBlockTreeDB::ReadTimestampBlockIndex(const uint256 &hash, unsigned int &ltimestamp) {

    CTimestampBlockIndexValue lts;
    if (!Read(std::make_pair(DB_BLOCKHASHINDEX, hash), lts)) {
        return false;
    }

    ltimestamp = lts.ltimestamp;
    return true;
}

bool CBlockTreeDB::WriteBlockBalancesIndex(const uint256 &key, const BlockBalances &value)
{
    CDBBatch batch(*this);
    batch.Write(std::make_pair(DB_BALANCESINDEX, key), value);
    return WriteBatch(batch);
}

bool CBlockTreeDB::ReadBlockBalancesIndex(const uint256 &key, BlockBalances &value)
{
    return Read(std::make_pair(DB_BALANCESINDEX, key), value);
}

bool CBlockTreeDB::WriteFlag(const std::string &name, bool fValue) {
    return Write(std::make_pair(DB_FLAG, name), fValue ? '1' : '0');
}

bool CBlockTreeDB::ReadFlag(const std::string &name, bool &fValue) {
    char ch;
    if (!Read(std::make_pair(DB_FLAG, name), ch))
        return false;
    fValue = ch == '1';
    return true;
}

bool CBlockTreeDB::LoadBlockIndexGuts(const Consensus::Params& consensusParams, std::function<CBlockIndex*(const uint256&)> insertBlockIndex)
{
    std::unique_ptr<CDBIterator> pcursor(NewIterator());

    pcursor->Seek(std::make_pair(DB_BLOCK_INDEX, uint256()));

    // Load m_block_index
    while (pcursor->Valid()) {
        if (ShutdownRequested()) return false;
        std::pair<char, uint256> key;
        if (pcursor->GetKey(key) && key.first == DB_BLOCK_INDEX) {
            CDiskBlockIndex diskindex;
            if (pcursor->GetValue(diskindex)) {
                // Construct block index object
                CBlockIndex* pindexNew  = insertBlockIndex(diskindex.GetBlockHash());
                pindexNew->pprev                    = insertBlockIndex(diskindex.hashPrev);
                pindexNew->nHeight                  = diskindex.nHeight;
                pindexNew->nFile                    = diskindex.nFile;
                pindexNew->nDataPos                 = diskindex.nDataPos;
                pindexNew->nUndoPos                 = diskindex.nUndoPos;
                pindexNew->nVersion                 = diskindex.nVersion;
                pindexNew->hashMerkleRoot           = diskindex.hashMerkleRoot;
                pindexNew->hashWitnessMerkleRoot    = diskindex.hashWitnessMerkleRoot;
                pindexNew->nTime                    = diskindex.nTime;
                pindexNew->nBits                    = diskindex.nBits;
                pindexNew->nNonce                   = diskindex.nNonce;
                pindexNew->nStatus                  = diskindex.nStatus;
                pindexNew->nTx                      = diskindex.nTx;

                pindexNew->nFlags                   = diskindex.nFlags & (uint32_t)~BLOCK_DELAYED;
                pindexNew->bnStakeModifier          = diskindex.bnStakeModifier;
                pindexNew->prevoutStake             = diskindex.prevoutStake;
                //pindexNew->hashProof                = diskindex.hashProof;

                pindexNew->nMoneySupply             = diskindex.nMoneySupply;
                pindexNew->nAnonOutputs             = diskindex.nAnonOutputs;


                if (pindexNew->nHeight == 0
                    && pindexNew->GetBlockHash() != Params().GetConsensus().hashGenesisBlock)
                    return error("LoadBlockIndex(): Genesis block hash incorrect: %s", pindexNew->ToString());

                if (fParticlMode) {
                    // only CheckProofOfWork for genesis blocks
                    if (diskindex.hashPrev.IsNull() && !CheckProofOfWork(pindexNew->GetBlockHash(),
                        pindexNew->nBits, Params().GetConsensus(), 0, Params().GetLastImportHeight()))
                        return error("%s: CheckProofOfWork failed: %s", __func__, pindexNew->ToString());
                } else
                if (!CheckProofOfWork(pindexNew->GetBlockHash(), pindexNew->nBits, Params().GetConsensus())) {
                    return error("%s: CheckProofOfWork failed: %s", __func__, pindexNew->ToString());
                }

                pcursor->Next();
            } else {
                return error("%s: failed to read value", __func__);
            }
        } else {
            break;
        }
    }

    return true;
}

bool CBlockTreeDB::ReadRCTOutput(int64_t i, CAnonOutput &ao)
{
    return Read(std::make_pair(DB_RCTOUTPUT, i), ao);
};

bool CBlockTreeDB::WriteRCTOutput(int64_t i, const CAnonOutput &ao)
{
    CDBBatch batch(*this);
    batch.Write(std::make_pair(DB_RCTOUTPUT, i), ao);
    return WriteBatch(batch);
};

bool CBlockTreeDB::EraseRCTOutput(int64_t i)
{
    CDBBatch batch(*this);
    batch.Erase(std::make_pair(DB_RCTOUTPUT, i));
    return WriteBatch(batch);
};


bool CBlockTreeDB::ReadRCTOutputLink(const CCmpPubKey &pk, int64_t &i)
{
    return Read(std::make_pair(DB_RCTOUTPUT_LINK, pk), i);
};

bool CBlockTreeDB::WriteRCTOutputLink(const CCmpPubKey &pk, int64_t i)
{
    CDBBatch batch(*this);
    batch.Write(std::make_pair(DB_RCTOUTPUT_LINK, pk), i);
    return WriteBatch(batch);
};

bool CBlockTreeDB::EraseRCTOutputLink(const CCmpPubKey &pk)
{
    CDBBatch batch(*this);
    batch.Erase(std::make_pair(DB_RCTOUTPUT_LINK, pk));
    return WriteBatch(batch);
};

bool CBlockTreeDB::ReadRCTKeyImage(const CCmpPubKey &ki, CAnonKeyImageInfo &data)
{
    // Versions before 0.19.2.15 store only the txid
    CDataStream ssValue(SER_DISK, CLIENT_VERSION);
    if (!ReadStream(std::make_pair(DB_RCTKEYIMAGE, ki), ssValue)) {
        return false;
    }
    try {
        if (ssValue.size() < 36) {
            ssValue >> data.txid;
            data.height = -1; // unset
        } else {
            ssValue >> data;
        }
    } catch (const std::exception&) {
        return false;
    }
    return true;
};

bool CBlockTreeDB::EraseRCTKeyImage(const CCmpPubKey &ki)
{
    CDBBatch batch(*this);
    batch.Erase(std::make_pair(DB_RCTKEYIMAGE, ki));
    return WriteBatch(batch);
};

bool CBlockTreeDB::EraseRCTKeyImagesAfterHeight(int height)
{
    CDBBatch batch(*this);
    size_t total = 0, removing = 0;
    std::unique_ptr<CDBIterator> pcursor(NewIterator());
    pcursor->Seek(std::make_pair(DB_RCTKEYIMAGE, CCmpPubKey()));

    while (pcursor->Valid()) {
        if (ShutdownRequested()) return false;
        std::pair<char, CCmpPubKey> key;
        if (pcursor->GetKey(key) && key.first == DB_RCTKEYIMAGE) {
            CAnonKeyImageInfo ki_data;
            total++;
            if (pcursor->GetValueSize() >= 36) {
                if (pcursor->GetValue(ki_data)) {
                    if (height < ki_data.height) {
                        removing++;
                        batch.Erase(std::make_pair(DB_RCTKEYIMAGE, key.second));
                    }
                } else {
                    return error("%s: failed to read value", __func__);
                }
            }
            pcursor->Next();
        } else {
            break;
        }
    }

    LogPrintf("Removing %d key images after height %d, total %d.\n", removing, height, total);
    if (removing < 1) {
        return true;
    }
    return WriteBatch(batch);
};

bool CBlockTreeDB::ReadSpentCache(const COutPoint &outpoint, SpentCoin &coin)
{
    return Read(std::make_pair(DB_SPENTCACHE, outpoint), coin);
};

bool CBlockTreeDB::EraseSpentCache(const COutPoint &outpoint)
{
    CDBBatch batch(*this);
    batch.Erase(std::make_pair(DB_SPENTCACHE, outpoint));
    return WriteBatch(batch);
};

bool CBlockTreeDB::EraseRewardTrackerUndo(int nHeight)
{
    CDBBatch batch(*this);
    batch.Erase(std::make_pair(DB_TRACKER_INPUTS_UNDO, nHeight));
    batch.Erase(std::make_pair(DB_TRACKER_OUTPUTS_UNDO, nHeight));
    return WriteBatch(batch);
}

bool CBlockTreeDB::ReadRewardTrackerUndo(ColdRewardUndo& rewardUndo, int nHeight)
{

    const std::unique_ptr<CDBIterator> pcursor(pblocktree->NewIterator());

    pcursor->Seek(std::make_pair(DB_TRACKER_INPUTS_UNDO, std::vector<std::pair<AddressType, CAmount>>()));

    while (pcursor->Valid()) {
        if (ShutdownRequested()) {
            return false;
        }
        std::pair<char, int> key;
        if (pcursor->GetKey(key) && key.first == DB_TRACKER_INPUTS_UNDO) {
            std::vector<std::pair<AddressType, CAmount>> inputs;
            if (pcursor->GetValue(inputs)) {
                rewardUndo.inputs[key.second] = std::move(inputs);
            }
            pcursor->Next();
        } else {
            break;
        }
    }

    pcursor->Seek(std::make_pair(DB_TRACKER_OUTPUTS_UNDO, std::vector<std::pair<AddressType, CAmount>>()));

    while (pcursor->Valid()) {
        if (ShutdownRequested()) {
            return false;
        }
        std::pair<char, int> key;
        if (pcursor->GetKey(key) && key.first == DB_TRACKER_OUTPUTS_UNDO) {
            std::vector<std::pair<AddressType, CAmount>> outputs;
            if (pcursor->GetValue(outputs)) {
                rewardUndo.outputs[key.second] = std::move(outputs);
            }
            pcursor->Next();
        } else {
            break;
        }
    }
    return true;
}

bool CBlockTreeDB::WriteRewardTrackerUndo(const ColdRewardUndo& rewardUndo)
{
    CDBBatch batch(*this);

    for (const auto& inputs: rewardUndo.inputs) {
        batch.Write(std::make_pair(DB_TRACKER_INPUTS_UNDO, inputs.first), inputs.second);
    }

    for (const auto& outputs: rewardUndo.outputs) {
        batch.Write(std::make_pair(DB_TRACKER_OUTPUTS_UNDO, outputs.first), outputs.second);
    }

    return WriteBatch(batch);
}

bool CBlockTreeDB::WriteLastTrackedHeight(std::int64_t lastHeight) {
    CDBBatch batch(*this);
    LogPrintf("%s Writting last tracked height %d\n", __func__, lastHeight);

    batch.Write(std::make_pair(DB_LAST_TRACKED_HEIGHT, 0), lastHeight);
    return WriteBatch(batch);
}

bool CBlockTreeDB::ReadLastTrackedHeight(std::int64_t& rv) {
    return Read(std::make_pair(DB_LAST_TRACKED_HEIGHT, 0), rv);
}

bool CBlockTreeDB::EraseLastTrackedHeight() {
    CDBBatch batch(*this);
    LogPrintf("%s Erasing last tracked height\n", __func__);
    batch.Erase(std::make_pair(DB_LAST_TRACKED_HEIGHT, 0));
    return WriteBatch(batch);
}


bool CCoinsViewDB::Upgrade()
{
    // TODO
    return true;
};

/* TODO: force reindex
namespace {

//! Legacy class to deserialize pre-pertxout database entries without reindex.
class CCoins
{
public:
    //! whether transaction is a coinbase
    bool fCoinBase;

    //! unspent transaction outputs; spent outputs are .IsNull(); spent outputs at the end of the array are dropped
    std::vector<CTxOut> vout;

    //! at which height this transaction was included in the active block chain
    int nHeight;

    //! empty constructor
    CCoins() : fCoinBase(false), vout(0), nHeight(0) { }

    template<typename Stream>
    void Unserialize(Stream &s) {
        unsigned int nCode = 0;
        // version
        unsigned int nVersionDummy;
        ::Unserialize(s, VARINT(nVersionDummy));
        // header code
        ::Unserialize(s, VARINT(nCode));
        fCoinBase = nCode & 1;
        std::vector<bool> vAvail(2, false);
        vAvail[0] = (nCode & 2) != 0;
        vAvail[1] = (nCode & 4) != 0;
        unsigned int nMaskCode = (nCode / 8) + ((nCode & 6) != 0 ? 0 : 1);
        // spentness bitmask
        while (nMaskCode > 0) {
            unsigned char chAvail = 0;
            ::Unserialize(s, chAvail);
            for (unsigned int p = 0; p < 8; p++) {
                bool f = (chAvail & (1 << p)) != 0;
                vAvail.push_back(f);
            }
            if (chAvail != 0)
                nMaskCode--;
        }
        // txouts themself
        vout.assign(vAvail.size(), CTxOut());
        for (unsigned int i = 0; i < vAvail.size(); i++) {
            if (vAvail[i])
                ::Unserialize(s, Using<TxOutCompression>(vout[i]));
        }
        // coinbase height
        ::Unserialize(s, VARINT_MODE(nHeight, VarIntMode::NONNEGATIVE_SIGNED));
    }
};

}

Upgrade the database from older formats.
 *
 * Currently implemented: from the per-tx utxo model (0.8..0.14.x) to per-txout.

bool CCoinsViewDB::Upgrade() {
    std::unique_ptr<CDBIterator> pcursor(m_db->NewIterator());
    pcursor->Seek(std::make_pair(DB_COINS, uint256()));
    if (!pcursor->Valid()) {
        return true;
    }

    int64_t count = 0;
    LogPrintf("Upgrading utxo-set database...\n");
    LogPrintf("[0%%]...\n"); / Continued /
    uiInterface.ShowProgress(_("Upgrading UTXO database").translated, 0, true);
    size_t batch_size = 1 << 24;
    CDBBatch batch(*m_db);
    int reportDone = 0;
    std::pair<unsigned char, uint256> key;
    std::pair<unsigned char, uint256> prev_key = {DB_COINS, uint256()};
    while (pcursor->Valid()) {
        if (ShutdownRequested()) {
            break;
        }
        if (pcursor->GetKey(key) && key.first == DB_COINS) {
            if (count++ % 256 == 0) {
                uint32_t high = 0x100 * *key.second.begin() + *(key.second.begin() + 1);
                int percentageDone = (int)(high * 100.0 / 65536.0 + 0.5);
                uiInterface.ShowProgress(_("Upgrading UTXO database").translated, percentageDone, true);
                if (reportDone < percentageDone/10) {
                    // report max. every 10% step
                    LogPrintf("[%d%%]...\n", percentageDone); / Continued /
                    reportDone = percentageDone/10;
                }
            }
            CCoins old_coins;
            if (!pcursor->GetValue(old_coins)) {
                return error("%s: cannot parse CCoins record", __func__);
            }
            COutPoint outpoint(key.second, 0);
            for (size_t i = 0; i < old_coins.vout.size(); ++i) {
                if (!old_coins.vout[i].IsNull() && !old_coins.vout[i].scriptPubKey.IsUnspendable()) {
                    Coin newcoin(std::move(old_coins.vout[i]), old_coins.nHeight, old_coins.fCoinBase);
                    outpoint.n = i;
                    CoinEntry entry(&outpoint);
                    batch.Write(entry, newcoin);
                }
            }
            batch.Erase(key);
            if (batch.SizeEstimate() > batch_size) {
                m_db->WriteBatch(batch);
                batch.Clear();
                m_db->CompactRange(prev_key, key);
                prev_key = key;
            }
            pcursor->Next();
        } else {
            break;
        }
    }
    m_db->WriteBatch(batch);
    m_db->CompactRange({DB_COINS, uint256()}, key);
    uiInterface.ShowProgress("", 100, false);
    LogPrintf("[%s].\n", ShutdownRequested() ? "CANCELLED" : "DONE");
    return !ShutdownRequested();
}
*/
